/**
 * \file
 *
 * \author	Georg Hopp
 *
 * \copyright
 * Copyright © 2014 Georg Hopp
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stdarg.h>
#include <stdint.h>

#include <sys/types.h>

#include "trbase.h"
#include "trio.h"

#include "tr/comm_end_point.h"
#include "tr/protocol.h"
#include "tr/interface/comm_end_point.h"

static
int
commEndPointCtor(void * _this, va_list * params)
{
	TR_CommEndPoint this = _this;

	this->transport       = va_arg(*params, TR_Socket);
	this->protocol        = va_arg(*params, TR_Protocol);
	this->read_chunk_size = va_arg(*params, int);
	this->do_close        = 0;
	this->write_buffer    = TR_new(TR_Queue);

	return 0;
}

static
void
commEndPointDtor(void * _this)
{
	TR_CommEndPoint this = _this;

	TR_delete(this->transport);
	TR_delete(this->write_buffer);
}

static
unsigned long
commEndPointGetHash(void * _this)
{
	return (unsigned long)((TR_CommEndPoint)_this)->transport->handle;
}

static
void
commEndPointHandleDouble(void * _current, void * _new)
{
	TR_CommEndPoint current = _current;
	TR_CommEndPoint new     = _new;

	// add will delete _new after this function is processed, so it's
	// neccessary to cleanup _current and reinit it with whatever was
	// in _new.
	// This will only be called if _current and _new are different.
	commEndPointDtor(current);

	current->transport       = new->transport;
	current->protocol        = new->protocol;
	current->read_chunk_size = new->read_chunk_size;
	current->do_close        = new->do_close;
	current->write_buffer    = new->write_buffer;
}

static
void
commEndPointCvInit(TR_class_ptr cls)
{
	TR_EVENT_CREATE(cls, TR_CEP_EVENT_DO_READ);
	TR_EVENT_CREATE(cls, TR_CEP_EVENT_DO_WRITE);
	TR_EVENT_CREATE(cls, TR_CEP_EVENT_READ_BLOCK);
	TR_EVENT_CREATE(cls, TR_CEP_EVENT_WRITE_BLOCK);
	TR_EVENT_CREATE(cls, TR_CEP_EVENT_WBUF_FULL);
	TR_EVENT_CREATE(cls, TR_CEP_EVENT_NEW_DATA);
	TR_EVENT_CREATE(cls, TR_CEP_EVENT_NEW_MSG);
	TR_EVENT_CREATE(cls, TR_CEP_EVENT_MSG_READY);
	TR_EVENT_CREATE(cls, TR_CEP_EVENT_DATA_READY);
	TR_EVENT_CREATE(cls, TR_CEP_EVENT_DATA_END);
	TR_EVENT_CREATE(cls, TR_CEP_EVENT_SHUT_READ);
	TR_EVENT_CREATE(cls, TR_CEP_EVENT_SHUT_WRITE);
	TR_EVENT_CREATE(cls, TR_CEP_EVENT_CLOSE);
}

const char * TR_cepEventStrings[] = {
	"TR_CEP_EVENT_DO_READ",
	"TR_CEP_EVENT_DO_WRITE",
	"TR_CEP_EVENT_READ_BLOCK",
	"TR_CEP_EVENT_WRITE_BLOCK",
	"TR_CEP_EVENT_WBUF_FULL",
	"TR_CEP_EVENT_NEW_DATA",
	"TR_CEP_EVENT_NEW_MSG",
	"TR_CEP_EVENT_MSG_READY",
	"TR_CEP_EVENT_DATA_READY",
	"TR_CEP_EVENT_DATA_END",
	"TR_CEP_EVENT_SHUT_READ",
	"TR_CEP_EVENT_SHUT_WRITE",
	"TR_CEP_EVENT_CLOSE",
};

intptr_t comm_end_point_events[TR_CEP_EVENT_MAX + 1];
TR_INIT_IFACE(TR_Class, commEndPointCtor, commEndPointDtor, NULL);
TR_INIT_IFACE(TR_CommEndPoint, NULL, NULL);
TR_INIT_IFACE(TR_Hashable, commEndPointGetHash, commEndPointHandleDouble);
TR_CREATE_CLASS(
		TR_CommEndPoint,
		TR_EventSubject,
		commEndPointCvInit,
		TR_IF(TR_Class),
		TR_IF(TR_CommEndPoint),
		TR_IF(TR_Hashable)) = {
	{
		TR_cepEventStrings,
		TR_CEP_EVENT_MAX + 1,
		comm_end_point_events
	}
};

// vim: set ts=4 sw=4:
