/**
 * \file
 *
 * \author	Georg Hopp
 *
 * \copyright
 * Copyright © 2014 Georg Hopp
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stdarg.h>
#include <stdint.h>

#include <sys/types.h>

#include "trbase.h"
#include "trio.h"
#include "trdata.h"

#include "tr/connection.h"
#include "tr/interface/protocol.h"
#include "tr/interface/comm_end_point.h"

static
int
connectionCtor(void * _this, va_list * params)
{
	TR_Connection this = _this;

	TR_PARENTCALL(TR_Connection, _this, TR_Class, ctor, params);
	this->current_message = NULL;

	return 0;
}

static
void
connectionDtor(void * _this)
{
	TR_Connection this = _this;

	TR_delete(this->current_message);
	TR_PARENTCALL(TR_Connection, _this, TR_Class, dtor);
}

static
TR_ProtoMessage
connectionNextMessage(void * _this, TR_RemoteData * data)
{
	TR_Connection   this        = _this;
	TR_CommEndPoint comm        = _this;
	TR_ProtoMessage ret_message = NULL;
	TR_RemoteData   new_data    = NULL;

	if (*data) {
		if (! this->current_message || this->current_message->ready)
		{
			this->current_message =
				TR_protoCreateMessage(comm->protocol, (*data)->remote);
		}

		/*
		 * This will return NULL if all data was consumed or a new TR_RemoteData
		 * with the remaining data.
		 * In other words, the protocol implementatio is completely responsible
		 * for the data management here. It can decide to copy the data, but
		 * then has to free the given TR_RemoteData by itself. This gives most
		 * flexibility when writing a protocol.
		 */
		new_data = TR_protoParse(comm->protocol, this->current_message, *data);

		/**
		 * We define that the only valid reason for a protocol parser to not
		 * consume all data is, that the current message is complete.
		 * When a parser returns a not completely consumed data then we first
		 * check if the current message is ready. If it is we create a
		 * new data object from the remaining data and return it to the caller
		 * along with the message. The caller (which is the protocol handler)
		 * can then call this again with the remaining data.
		 * If the message is not ready we drop the data silently because there
		 * is either wrong data or a bug in the parser and the data will never
		 * be consumed correctly.
		 * INFO: Usually we do not free data here at all. We leave this to the
		 * protocol implementation. The protocol might take the data without
		 * copying it at all or if it copies it is responsible for the free too.
		 * Only if we got a wrong behaviour of the protocol we free the data
		 * ourself.
		 * IMPORTANT: The protocol should never free the data when it does not
		 * consume it completely.
		 * IMPORTANT: To keep this maintainable we must write a log message here
		 * when we drop data. This message should be a WARNING as the protocol
		 * might want to drop data intentionally... (probably this is not
		 * true and we should make it an ERROR).
		 */
		if (this->current_message->ready) {
			ret_message = this->current_message;
			this->current_message = NULL;
		} else {
			if (new_data) {
				TR_delete(*data);
				TR_delete(new_data);
				TR_loggerLog(
						TR_logger,
						TR_LOGGER_WARNING,
						"Drop data not consumed by protocol.");
			}
		}

		*data = new_data;
	}

	return ret_message;
}

static
size_t
connectionCompose(void * _this, TR_ProtoMessage message)
{
	TR_RemoteData data =
		TR_protoCompose(((TR_CommEndPoint)_this)->protocol, message);

	if (! data) {
		return FALSE;
	}

	TR_queuePut(((TR_CommEndPoint)_this)->write_buffer, data);
	return ((TR_SizedData)data)->size;
}

static
void
connectionCvInit(TR_class_ptr cls)
{
	TR_EVENT_CREATE(cls, TR_CON_EVENT_NEW_CON);
}

const char * TR_connectionEventStrings[] = {
	"TR_CON_EVENT_NEW_CON"
};

intptr_t connection_events[TR_CON_EVENT_MAX + 1];
TR_INIT_IFACE(TR_Class, connectionCtor, connectionDtor, NULL);
TR_INIT_IFACE(
		TR_CommEndPoint,
		connectionNextMessage,
		connectionCompose);
TR_CREATE_CLASS(
		TR_Connection,
		TR_CommEndPoint,
		connectionCvInit,
		TR_IF(TR_Class),
		TR_IF(TR_CommEndPoint)) = {
	{{
		 TR_connectionEventStrings,
		 TR_CON_EVENT_MAX + 1,
		 connection_events
	}}
};

// vim: set ts=4 sw=4:
