/**
 * \file
 *
 * \author	Georg Hopp
 *
 * \copyright
 * Copyright © 2014 Georg Hopp
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <unistd.h>
#include <inttypes.h>

#include "trbase.h"
#include "trevent.h"

#include "tr/protocol.h"
#include "tr/connection.h"
#include "tr/protocol_handler.h"
#include "tr/proto_message.h"
#include "tr/comm_end_point.h"
#include "tr/interface/comm_end_point.h"

static
int
protocolHandlerCtor(void * _this, va_list * params)
{
	TR_PARENTCALL(TR_ProtocolHandler, _this, TR_Class, ctor, params);

	return 0;
}

static void protocolHandlerDtor(void * _this) {}

static
TR_EventDone
protocolHandlerParse(void * _this, TR_Event event)
{
	/**
	 * TODO No upgrade for now. Add it later on.
	 */
	TR_CommEndPoint endpoint = (TR_CommEndPoint)event->subject;
	TR_RemoteData   data     = event->data;
	TR_ProtoMessage message;

	while ((message = TR_cepNextMessage(endpoint, &data))) {
		if (! TR_eventHandlerIssueEvent(
				(TR_EventHandler)_this,
				TR_eventSubjectEmit(
					event->subject,
					TR_CEP_EVENT_NEW_MSG,
					message))) {
			TR_delete(message);
		} else {
			if (message->close) {
				// also check that we are a response. Well this is
				// how it is done in the python code...
				TR_cepSetClose(endpoint);
			}
		}
	}

	return TR_EVENT_DONE;
}

static
TR_EventDone
protocolHandlerCompose(void * _this, TR_Event event)
{
	TR_CommEndPoint endpoint = (TR_CommEndPoint)event->subject;
	TR_ProtoMessage message  = (TR_ProtoMessage)event->data;
	size_t          message_size;

	if (message->close) {
		// also check that we are a response. Well this is how it is done
		// in the python code...
		TR_cepSetClose(endpoint);
	}

	if ((message_size = TR_cepCompose(endpoint, message))) {
		endpoint->write_buffer_size += message_size;

		if (endpoint->write_buffer_size >= CEP_WRITE_BUFFER_THRESHOLD) {
			TR_eventHandlerIssueEvent(
					(TR_EventHandler)_this,
					TR_eventSubjectEmit(
						event->subject,
						TR_CEP_EVENT_WBUF_FULL,
						NULL));
		}

		if (TR_queueSize(endpoint->write_buffer) >= 1) {
			TR_eventHandlerIssueEvent(
					(TR_EventHandler)_this,
					TR_eventSubjectEmit(
						event->subject,
						TR_CEP_EVENT_DATA_READY,
						NULL));
		}
	}
	TR_delete(message);

	return TR_EVENT_DONE;
}

TR_EventDone
protocolHandlerUpgrade(void * _this, TR_Event event)
{
	return TR_EVENT_DONE;
}

static
void
protocolHandlerCvInit(TR_class_ptr cls)
{
	TR_EVENT_HANDLER_SET_METHOD(
			cls,
			TR_CommEndPoint,
			TR_CEP_EVENT_NEW_DATA,
			protocolHandlerParse);
	TR_EVENT_HANDLER_SET_METHOD(
			cls,
			TR_CommEndPoint,
			TR_CEP_EVENT_MSG_READY,
			protocolHandlerCompose);
//	TR_EVENT_HANDLER_SET_METHOD(
//			cls,
//			TR_CommEndPoint,
//			TR_CEP_EVENT_UPGRADE,
//			protocolHandlerUpgrade);
}

TR_INIT_HANDLER(TR_ProtocolHandler);
TR_INIT_IFACE(TR_Class, protocolHandlerCtor, protocolHandlerDtor, NULL);
TR_CREATE_CLASS(
		TR_ProtocolHandler,
		TR_EventHandler,
		protocolHandlerCvInit,
		TR_IF(TR_Class)) = {
	{ TR_HANDLER_CVARS(TR_ProtocolHandler) }
};

// vim: set ts=4 sw=4:
